# coding=utf-8
from dateutil.easter import EASTER_WESTERN

from holidata.utils import SmartDayArrow
from .holidays import Holiday, Locale


#
# Information taken from various government websites around 2020-04-16
#   https://www.canada.ca/fr/emploi-developpement-social/services/normes-travail/rapports/jours-feries.html
#   http://www4.gouv.qc.ca/FR/Portail/Citoyens/Evenements/travailleur-avec-salaire/Pages/jours-feries-chomes-payes.aspx
#   https://www.ontario.ca/document/your-guide-employment-standards-act-0/public-holidays#section-2
#   https://www.cfib-fcei.ca/en/tools-resources/paying-employees-public-holidays-newfoundland-labrador
#   https://www.princeedwardisland.ca/en/information/economic-growth-tourism-and-culture/paid-holidays
#   https://www2.gnb.ca/content/dam/gnb/Departments/petl-epft/PDF/es/FactSheets/PublicHolidaysVacation.pdf
#   https://www.gov.mb.ca/labour/standards/doc,gen-holidays-after-april-30-07,factsheet.html#q12
#   https://www.saskatchewan.ca/business/employment-standards/vacations-holidays-leaves-and-absences/public-statutory-holidays/list-of-saskatchewan-public-holidays
#
# Also those sites for some information
#   https://www.officeholidays.com/holidays/canada/canada-remembrance-day
#   https://www.timeanddate.com/holidays/canada/family-day
#   https://fr.wikipedia.org/wiki/F%C3%AAtes_et_jours_f%C3%A9ri%C3%A9s_au_Canada
#
# I have not checked every province and territory website, but the wikipedia
# summary has been true for everything that I have checked, although it seems
# to be considering more holidays than the bare minimum and also counts holidays
# that are not mandated but usually observed (e.g. St-Patrick's Day in NL which
# is not statutory, but is given to government employees).
#


class fr_CA(Locale):
    """
    01-01: [NF] Jour de l'An
    06-24: [QC] [F] Fête Nationale
    07-01: [NF] Fête du Canada
    11-11: [AB,BC,NB,NL,NT] [F] Jour du Souvenir
    12-25: [NRF] Jour de Noël
    12-26: [NRF] Lendemain de Noël
    2 days before Easter: [NRV] Vendredi Saint
    1 day after Easter: [AB,PE,QC] [RV] Lundi de Pâques
    3. monday in February: [AB,ON,SK,NB] [V] Fête de la Famille
    3. monday in February: [MB] [V] Journée Louis Riel
    3. monday in February: [PE] [V] Fête des Insulaires
    1. monday in August: [NT,NU] [V] Premier lundi d'août
    1. monday in August: [AB] [V] Fête du Patrimoine
    1. monday in August: [SK] [V] Fête de la Saskatchewan
    1. monday in August: [NS] [V] Jour de la Fondation
    1. monday in August: [NB] [V] Jour du Nouveau-Brunswick
    1. monday in September: [NV] Fête du Travail
    2. monday in October: [AB,BC,MB,NL,ON,QC,SK,NT,NU,YT] [V] Jour de l'Action de grâce
    """

    locale = "fr-CA"
    easter_type = EASTER_WESTERN

    def holiday_journee_nationale_des_patriotes(self):
        return [Holiday(
            locale=self.locale,
            region="QC",
            date=SmartDayArrow(self.year, 5, 25).shift_to_weekday(
                "monday", order=1, reverse=True, including=False),
            description="Journée Nationale des Patriotes",
            flags="V",
            notes="")]

    def holiday_fete_de_la_reine_victoria(self):
        return [Holiday(
            locale=self.locale,
            region=region,
            date=SmartDayArrow(self.year, 5, 25).shift_to_weekday(
                "monday", order=1, reverse=True, including=False),
            description="Fête de la Reine Victoria",
            flags="V",
            notes=""
        ) for region in ["AB", "BC", "MB", "NS", "ON", "SK", "NT", "NU", "YT"]
        ]



import re

import pytest

from holidata import Locale, Country
from tests import HOLIDATA_YEAR_MAX


@pytest.fixture(params=range(2011, HOLIDATA_YEAR_MAX))
def year(request):
    return request.param


@pytest.fixture(params=Locale.plugins)
def locale(request, year):
    return request.param(year)


@pytest.fixture()
def holidays(locale):
    return locale.holidays


@pytest.fixture(params=Country.plugins)
def country(request):
    return request.param()


def test_country_should_be_constructable(country):
    pass


def test_holiday_should_not_be_of_type_national_if_region_defined(holidays):
    for holiday in holidays:
        if holiday.region == "":
            assert "N" in holiday.flags, f"Holiday '{holiday.description}' ({holiday.date.strftime('%Y-%m-%d')}) in locale {holiday.locale} must have flag 'N': it has no regions defined"


def test_holiday_should_be_of_type_national_if_no_region_defined(holidays):
    for holiday in holidays:
        if holiday.region != "":
            assert "N" not in holiday.flags, f"Holiday '{holiday.description}' ({holiday.date.strftime('%Y-%m-%d')}) in locale {holiday.locale} must not have flag 'N': it has regions defined"


def test_holiday_should_be_of_type_either_fixed_or_variable(holidays):
    for holiday in holidays:
        date_is_fixed = "F" in holiday.flags
        date_is_variable = "V" in holiday.flags

        assert not (date_is_variable and date_is_fixed), f"Holiday '{holiday.description}' ({holiday.date.strftime('%Y-%m-%d')}) in locale {holiday.locale} must not have both flags 'F' and 'V'"
        assert (date_is_variable or date_is_fixed), f"Holiday '{holiday.description}' ({holiday.date.strftime('%Y-%m-%d')}) in locale {holiday.locale} must have either flag 'F' or 'V'"


def test_holiday_flags_should_be_in_the_correct_order(holidays):
    for holiday in holidays:
        match = re.search(r"^N?R?[FV]?$", f"{holiday.flags}")

        assert match is not None, f"Flags for holiday '{holiday.description}' ({holiday.date.strftime('%Y-%m-%d')}) in locale {holiday.locale} are not in the correct order. Flags '{holiday.flags}' should match 'N?R?[FV]?'"



from .holidays import *
from .emitters import Emitter


def get_country_for(identifier):
    country_class = Country.get(identifier)

    if not country_class:
        raise ValueError(f"No plugin found for country id '{identifier}'!")

    return country_class()


def get_locale_class_for(identifier):
    locale_class = Locale.get(identifier)

    if not locale_class:
        raise ValueError(f"No plugin found for locale '{identifier}'!")

    return locale_class


def create_locale_for(country_id=None, lang_id=None, year=None):
    country_class = get_country_for(country_id)

    if lang_id is not None and lang_id.lower() not in country_class.languages:
        raise ValueError(f"Language '{lang_id}' is not defined for country '{country_class.id}'!")
    elif lang_id is None and country_class.default_lang is not None:
        lang_id = country_class.default_lang
    elif lang_id is None:
        raise ValueError(f"Country '{country_id}' has no default language specified! Choose one of [{', '.join(country_class.languages)}].")

    locale_class = get_locale_class_for(f"{lang_id}-{country_id}")

    return locale_class(year)


def create_emitter_for(identifier):
    emitter_class = Emitter.get(identifier)

    if not emitter_class:
        raise ValueError(f"Unsupported output format '{identifier}'!")

    return emitter_class()


def parse_year(year):
    try:
        return int(year)
    except ValueError:
        raise ValueError(f"Invalid year '{year}'! Has to be an integer.")


class Holidata:
    locale = None
    emitter = None

    def __init__(self, country=None, language=None, year=None, output=None):
        self.locale = create_locale_for(country_id=country, lang_id=language, year=parse_year(year))
        self.emitter = create_emitter_for(output)

    def __str__(self):
        return self.emitter.output(self.locale)



import re

from holidata.plugin import PluginMount
from holidata.utils import SmartDayArrow, month_reference, easter


class Holiday(object):
    """
    A sheer container for one holiday.
    """

    def __init__(self, locale, region, date, description, flags="", notes=""):
        self.locale = locale
        self.region = region
        self.date = date
        self.description = description
        self.flags = flags
        self.notes = notes

    def as_dict(self):
        return {
            "locale": self.locale,
            "region": self.region,
            "date": self.date.strftime("%Y-%m-%d"),
            "description": self.description,
            "type": self.flags,
            "notes": self.notes
        }


class Country(object, metaclass=PluginMount):
    """
    Represents holidays of a given country
    """
    id = None
    languages = []
    default_lang = None
    easter_type = None

    def __init__(self):
        if self.id is None:
            raise ValueError(f"Country '{self.__class__.__name__}' does not provide its id!")

        if not self.languages:
            raise ValueError(f"Country '{self.__class__.__name__}' does not list any languages!")

        if self.default_lang is not None and self.default_lang not in self.languages:
            raise ValueError(f"Country '{self.__class__.__name__}' does not list language '{self.default_lang}'!")

    @staticmethod
    def get(identifier):
        return Country.get_plugin(identifier, "id")


class Locale(object, metaclass=PluginMount):
    """
    Represents holidays in a given locale.
    """
    locale = None
    easter_type = None

    fixed_regex = re.compile(r"^\s*(?P<month>\d\d)-(?P<day>\d\d): "
                             r"(\[(?P<regions>[^]]+)\]\s+)?"
                             r"\[(?P<flags>[A-Z]*)\] (?P<description>.*)$", re.UNICODE)
    nth_weekday_regex = re.compile(r"^\s*(?P<order>\d+)\.(?P<last> last | )"
                                   r"(?P<weekday>[a-z]+) in (?P<month>[a-zA-Z]+):\s+"
                                   r"(\[(?P<regions>[^]]+)\]\s+)?"
                                   r"\[(?P<flags>[A-Z]*)\] (?P<description>.*)$", re.UNICODE)
    easter_shift_regex = re.compile(r"^\s*((?P<days>\d+) day(s)? (?P<direction>(before|after)) )?Easter:\s+"
                                    r"(\[(?P<regions>[^]]+)\]\s+)?"
                                    r"\[(?P<flags>[A-Z]*)\] (?P<description>.*)$", re.UNICODE)

    def __init__(self, year):
        if self.locale is None:
            raise ValueError(f"Locale {self.__class__.__name__} does not provide its locale")

        self.year = year

    @staticmethod
    def get(identifier):
        return Locale.get_plugin(identifier, "locale")

    @property
    def holidays(self):
        """
        Yield all the Holiday objects corresponding to the definitions in the
        self.__doc__ and also as given by the dynamic self.holiday_* methods.
        """
        # First process lines in the __doc__
        for line in self.__doc__.splitlines():
            # Skip empty lines
            if not line.strip():
                continue

            holidata = self._parse_holidata(line)

            if holidata is None:
                print(f"Following line could not be processed: '{line}'")
                continue

            for region in holidata["regions"]:
                yield Holiday(
                    locale=self.locale,
                    region=region,
                    date=holidata["date"],
                    description=holidata["description"],
                    flags=holidata["flags"],
                    notes=holidata["notes"],
                )

        # Second, call holiday functions
        for method in [getattr(self, func) for func in dir(self) if func.startswith("holiday_")]:
            holidays = method()

            for holiday in holidays:
                yield holiday

    def _parse_holidata(self, line):
        function_map = [
            (self.fixed_regex, self._date_from_fixed_reference),
            (self.nth_weekday_regex, self._date_from_weekday_reference),
            (self.easter_shift_regex, self._date_from_easter_reference),
        ]

        for reg_exp, create_date_from in function_map:
            m = reg_exp.search(line)
            if m is not None:
                return dict(regions=([x.strip() for x in m.group("regions").split(",")] if m.group("regions") is not None else [""]),
                            date=create_date_from(m),
                            description=m.group("description"),
                            flags=m.group("flags"),
                            notes="")

        return None

    def _date_from_fixed_reference(self, m):
        return SmartDayArrow(self.year, int(m.group("month")), int(m.group("day")))

    def _date_from_weekday_reference(self, m):
        return month_reference(self.year,
                               m.group("month"),
                               first=m.group("last").strip() == "") \
            .shift_to_weekday(m.group("weekday"),
                              order=int(m.group("order")),
                              reverse=m.group("last").strip() == "last",
                              including=True)

    def _date_from_easter_reference(self, m):
        if self.easter_type is None:
            raise ValueError(f"Locale {self.__class__.__name__} does not provide its easter type (WESTERN|ORTHODOX)")

        return easter(self.year, self.easter_type) \
            .shift(days=int((m.group("days")) if m.group("days") is not None else 0) * (1 if m.group("direction") == "after" else -1))
